==> pingvo-2.2.tar.bz2: pingvo-2.2/indexer.cc <==
#include <stdio.h>
#include <stdlib.h>

#ifdef __linux__
#include <unistd.h>
#include <sys/mman.h>
#else
#undef UNICODE
#include <windows.h>
#endif

#include <fcntl.h>
#include <string.h>
#include <q3cstring.h>
#include <qdom.h>
#include <qfile.h>
#include "indexer.h"

#include <algorithm>
using namespace std;

#ifndef __linux__
#include <stdarg.h>
int asprintf( char **s, char *fmt, ... ) {
	int rc, len = 1024;
	char *buf;
	va_list vl;
	
	va_start( vl, fmt );

	buf = (char*) malloc( len );
	for (;;) {
		rc = _vsnprintf( buf, len-1, fmt, vl );
		if ( rc >= 0 ) break;
		len += 1024;
		buf = (char*) realloc( buf, len );
	}
	*s = buf;
	return rc;
}
#define strncasecmp _strnicmp
#endif

char *IndexBuilder::words_start = 0;

IndexBuilder::IndexBuilder(): acount(0) {
  rootNode.word_number = -1;
  lookupTable = &rootNode.sub;
}

void IndexBuilder::build( const char *xmlname ) {
  char *base, *idxname, *wordname, *treename;
  char *idxname2, *wordname2;
  size_t len;
  FILE *f;
  
  base = strdup( xmlname );
  len = strlen(base) - 4 ;
  base[len] = 0;
  asprintf( &idxname, "%s.idx", base );
  asprintf( &wordname, "%s.words", base );
  asprintf( &treename, "%s.tree", base );

  puts( "indexing source xml" );
  build_index( xmlname, idxname );
  printf( "%d articles indexed\n", acount );
  puts( "building wordlist" );
  build_wordlist( xmlname, idxname, wordname );

  puts( "sorting wordlist" );
  asprintf( &idxname2, "%s.idx2", base );
  asprintf( &wordname2, "%s.words2", base );
  sort_words( idxname, idxname2, wordname );
  unlink( idxname );
  rename( idxname2, idxname );

  puts( "writing suffix tree" );
  f = fopen( treename, "wb" );
  if ( !f ) return;
//  ftruncate( fileno(f), 0 );
  write_tree( f, &rootNode );
  fclose( f );
}

void IndexBuilder::addToLookupTable( QString text, int word_num ) {
  unsigned int i, j;
  size_t sz;
  QChar c;
  vector <lookupEntry*> *dir;
  lookupEntry *ent=0;

  dir = lookupTable;
  for ( i = 0; i < text.length(); i++ ) {
        c = text.at(i);
        sz = dir->size();
        for ( j = 0; j < sz; j++ )
                if ( (*dir)[j]->c == c ) {
                        ent = (*dir)[j];
                        dir = &ent->sub;
                        break;
                }
        if ( j == sz ) break;
  }

  if ( i == text.length() ) {
        if ( ent ) {
		ent->word_number = word_num;
		ent->word = text;
	}
        return;
  }

  for ( ; i < text.length(); i++ ) {
        ent = new lookupEntry;
        ent->c = text.at(i);
		ent->word_number = -1;
        ent->word = QString::null;
        dir->push_back( ent );
        dir = &ent->sub;
  }

  ent->word_number = word_num;
  ent->word = text;
}

int IndexBuilder::build_index( const char *xmlname, const char *idxname ) {
  int state, offs, loffs, asize, bidx;
  char buf[80];
  int c;
  FILE *f, *f2;

  f = fopen( xmlname, "rb" );
  if ( !f ) {
  	perror( "build_index fopen(xmlname) failed" );
	return -1;
  }

  f2 = fopen( idxname, "wb" );
  if ( !f2 ) {
  	perror( "build_index fopen(idxname) failed" );
	return -1;
  }

  state = 0;
  acount = 0;
  offs = 0;
  while ( !feof(f) ) {
  	c = fgetc( f );
	if ( c == EOF  ) break;
	switch ( state ) {
	case 0:
		if ( c == '<' ) {
			loffs = offs;
			state = 1;
			bidx = 0;
		}
		break;
	case 1:	
		if ( c != '>' ) {
			if ( bidx < sizeof buf ) buf[bidx++] = c;
			break;
		}

		if ( !strncasecmp( buf, "entry", 5 ) ) state = 2;
		else state = 0;

		break;
	case 2:
		if ( c == '<' ) {
			state = 3;
			bidx = 0;
		}
		break;
	case 3:
		if ( c != '>' ) {
			if ( bidx < sizeof buf ) buf[bidx++] = c;
			break;
		}
		
		if ( strncasecmp( buf, "/entry", 6 ) ) {
			state = 2;
			break;
		}

		state = 0;
		asize = offs - loffs + 1;
		fwrite( &loffs, sizeof offs, 1, f2 );
		fwrite( &asize, sizeof asize, 1, f2 );
		acount++;
		break;
	}
	offs++;
  }
  fclose( f2 );
  fclose( f );

  return acount;
}

void IndexBuilder::build_wordlist( const char *xmlfile,
	const char *idxfile, const char *wordsfile ) {
  QDomDocument doc("entry");
  FILE *idx, *xml, *f2;
  size_t rc;
  int idxrec[2], word_num, idxpos;
  char *buf = 0;
  char bufsize = 0;
  word_desc wdesc;
  QString it_name;

  xml = fopen( xmlfile, "rb" );
  if ( !xml ) return;
  idx = fopen( idxfile, "rb" );
  if ( !idx ) {
  	fclose( xml );
  	return;
  }
  f2 = fopen( wordsfile, "wb" );
  if ( !f2 ) {
  	fclose( idx );
	fclose( xml );
	return;
  }

  fseek( f2, acount * 4, SEEK_SET );
  
  word_num = 0;
  words.clear();

  idxpos = 0;
  for (;;) {
  	char *word;

  	rc = fread( idxrec, 1, sizeof idxrec, idx );
	if ( rc != sizeof idxrec ) break;
	fseek( xml, idxrec[0], SEEK_SET );
	if ( idxrec[1] > bufsize ) {
		buf = (char*) realloc( buf, idxrec[1] );
		bufsize = idxrec[1];
	}
	fread( buf, 1, idxrec[1], xml );
	doc.setContent( Q3CString(buf) );
	it_name = doc.documentElement().attribute("word").simplifyWhiteSpace();
	word = strdup( it_name.utf8().data() );
	wdesc.dict_pos = word_num;
	wdesc.offs = ftell( f2 );
	words.push_back( wdesc );

	fseek( f2, word_num * 4, SEEK_SET );
	fwrite( &wdesc.offs, sizeof wdesc.offs, 1, f2 );
	fseek( f2, wdesc.offs, SEEK_SET );

	fwrite( word, 1, strlen(word)+1, f2 );
	free( word );

	word_num++;
  }

  if ( buf ) free( buf );
  fclose( f2 );
  fclose( idx );
  fclose( xml );
}

bool IndexBuilder::less_worddesc( const word_desc &w1, const word_desc &w2 ) {
  QString s1, s2;
  
  s1 = QString::fromUtf8( words_start + w1.offs );
  s2 = QString::fromUtf8( words_start + w2.offs );

  return s1 < s2;
}

#ifdef __linux__
void IndexBuilder::map_words( char *fname, char* &words_start, int& lenwords ) {
  int fdwords = open( fname, O_RDWR );
  lenwords = lseek( fdwords, 0, SEEK_END );
  words_start = (char*) mmap( 0, lenwords, PROT_READ|PROT_WRITE,
  	MAP_SHARED, fdwords, 0 );
  ::close( fdwords );
}

void IndexBuilder::unmap_words( char *words_start, int lenwords ) {
  munmap( words_start, lenwords );
}

void IndexBuilder::process_idx( char *idxfile, char *idxfile2 ) {
  int fdidx, lenidx, *idx_start;
  unsigned int i;

  fdidx = open( idxfile, O_RDONLY );
  lenidx = lseek( fdidx, 0, SEEK_END );
  idx_start = (int*) mmap( 0, lenidx, PROT_READ, MAP_SHARED, fdidx, 0 );
  close( fdidx );
  fdidx = open( idxfile2, O_RDWR|O_CREAT|O_TRUNC, 0644 );
  for ( i = 0; i < words.size(); i++ )
	write( fdidx, idx_start+(words[i].dict_pos << 1), sizeof(int) * 2 );
  munmap( idx_start, lenidx );
  close( fdidx );
}

#else

void IndexBuilder::map_words( char *file, char* &words_start, int& lenwords ) {
  HANDLE hFile, hMap;
  hFile = CreateFile( file, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
  if ( hFile == INVALID_HANDLE_VALUE ) {
	  puts( "aaa" );
  }
  lenwords = GetFileSize( hFile, NULL );
  hMap = CreateFileMapping( hFile, 0, PAGE_READWRITE, 0, 0, NULL );
  words_start = (char*) MapViewOfFile( hMap, FILE_MAP_READ|FILE_MAP_WRITE, 0, 0, 0 ); 
  CloseHandle( hFile );
  CloseHandle( hMap );
}

void IndexBuilder::unmap_words( char *words_start, int /* lenwords */ ) {
  UnmapViewOfFile( words_start );
}

void IndexBuilder::process_idx( char *idxfile, char *idxfile2 ) {
  HANDLE hFile, hMap;
  int *idx_start;
  unsigned int i;
  DWORD written;

  hFile = CreateFile( idxfile, GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
  hMap = CreateFileMapping( hFile, 0, PAGE_READONLY, 0, 0, NULL );
  idx_start = (int*) MapViewOfFile( hMap, FILE_MAP_READ, 0, 0, 0 );
  CloseHandle( hFile );
  CloseHandle( hMap );

  hFile = CreateFile( idxfile2, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
		NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
  for ( i = 0; i < words.size(); i++ )
	WriteFile( hFile, idx_start+(words[i].dict_pos << 1), sizeof(int) * 2, &written, 0 );
  CloseHandle( hFile );

  UnmapViewOfFile( idx_start );
}
#endif

void IndexBuilder::sort_words( char* idxfile, char* idxfile2,
	char* wordsfile ) {
  unsigned int i;
  int lenwords;
  QString word;

  map_words( wordsfile, words_start, lenwords );
  sort( words.begin(), words.end(), less_worddesc );
  process_idx( idxfile, idxfile2 );

  for ( i = 0; i < words.size(); i++ )
  	((int*)words_start)[i] = words[i].offs;

  puts( "building suffix tree" );
  for ( unsigned int word_num = 0; word_num < words.size(); word_num++ ) {
	word = QString::fromUtf8( words_start + words[word_num].offs );
	addToLookupTable( word, word_num );
  }

  unmap_words( words_start, lenwords );
}

void IndexBuilder::write_tree( FILE *f, lookupEntry *ent ) {
  unsigned int i;
  size_t subpos;
  int pos;
  lookupEntryExt ext;

  ext.c = ent->c.unicode();
  ext.word_number = ent->word_number;
  ext.dchilds_count = ent->sub.size();
  ext.childs_count = ent->childs_count;
  fwrite( &ext, sizeof ext, 1, f );
  pos = ftell( f );

  subpos = pos + ent->sub.size() * 4;
  fseek( f, subpos, SEEK_SET );
  for ( i = 0; i < ent->sub.size(); i++ ) {
	write_tree( f, ent->sub[i] );
	fseek( f, pos + i * 4, SEEK_SET );
	fwrite( &subpos, 4, 1, f );
	fseek( f, 0, SEEK_END );
	subpos = ftell( f );
  }
}

IndexBuilderDsl::IndexBuilderDsl() {
}

void IndexBuilderDsl::build( const char *datname0 ) {
  size_t namelen = strlen( datname0 );
  if ( namelen < 4 || strcmp( datname0 + namelen - 4, ".dat" ) ) return;

  char *base, *idxname2, *treename;
	  
  base = strdup( datname0 );
  base[namelen-4] = 0;

  dsldat = strdup( datname0 );
  asprintf( &dslidx, "%s.ddx", base );
  asprintf( &idxname, "%s.idx", base );
  asprintf( &idxname2, "%s.idx2", base );
  asprintf( &wordsname, "%s.words", base );
  asprintf( &treename, "%s.tree", base );

  acount = build_index();
  build_wordlist();

  sort_words( idxname, idxname2, wordsname );
  unlink( idxname );
  rename( idxname2, idxname );

  FILE *f = fopen( treename, "wb" );
  if ( !f ) return;
  write_tree( f, &rootNode );
  fclose( f );
}

int IndexBuilderDsl::build_index() {
  unsigned short reclen, asize16;
  unsigned char wordlen;
  unsigned int offs, loffs, asize;
  int acount, rc;
  FILE *f, *f1, *f2;

  f = fopen( dslidx, "rb" );
  if ( !f ) {
  	perror( "build_index fopen(dslidx) failed" );
	return -1;
  }

  f1 = fopen( dsldat, "rb" );
  if ( !f1 ) {
	perror( "build_index fopen(dsldat) failed" );
	return -1;
  }

  f2 = fopen( idxname, "wb+" );
  if ( !f2 ) {
  	perror( "build_index fopen(idxname) failed" );
	return -1;
  }

  offs = 0;
  acount = 0;
  while ( !feof(f) ) {
	rc = fread( &reclen, sizeof reclen, 1, f );
	if ( rc != 1 ) break;

	fread( &wordlen, sizeof wordlen, 1, f );
	wordlen <<= 1;
	offs += sizeof(reclen) + sizeof(wordlen) + wordlen;
	fseek( f, offs, SEEK_SET );
	fread( &loffs, sizeof loffs, 1, f );

	fseek( f1, loffs, SEEK_SET );
	fread( &asize16, sizeof asize16, 1, f1 );
	asize = asize16 * 2;

	loffs += sizeof(asize16);
	fwrite( &loffs, sizeof loffs, 1, f2 );
	fwrite( &asize, sizeof asize, 1, f2 );

	offs += sizeof(loffs) + 10;
	fseek( f, offs, SEEK_SET );
	acount++;
  }

  fclose( f );
  fclose( f1 );
  fclose( f2 );

  return acount;
}

void IndexBuilderDsl::build_wordlist() {
  FILE *f, *f1;
  int i, wlen2;
  unsigned int offs, woffs;
  unsigned short reclen;
  unsigned char wordlen;
  unsigned short buf[256];
  QString word;
  Q3CString wordUtf8;
  word_desc wdesc;

  f = fopen( dslidx, "rb" );
  f1 = fopen( wordsname, "wb" );

  words.clear();

  offs = 0;
  woffs = acount * sizeof woffs;
  for ( i = 0; i < acount; i++ ) {
	wdesc.dict_pos = i;
	wdesc.offs = woffs;
	words.push_back( wdesc );

	fseek( f1, i * sizeof woffs, SEEK_SET );
	fwrite( &woffs, sizeof woffs, 1, f1 );

	fseek( f1, woffs, SEEK_SET );

	fread( &reclen, sizeof reclen, 1, f );
	fread( &wordlen, sizeof wordlen, 1, f );
	wlen2 = wordlen << 1;
	fread( buf, wlen2, 1, f );
	buf[wordlen] = 0;
	word = QString::fromUcs2( buf );
	wordUtf8 = word.utf8();
	fwrite( wordUtf8.data(), wordUtf8.length()+1, 1, f1 );
	woffs += wordUtf8.length() + 1;

	offs += sizeof(reclen) + sizeof(wordlen) + wlen2 + 4 + 10;
	fseek( f, offs, SEEK_SET );
  }

  fclose( f );
  fclose( f1 );
}

==> pingvo-2.3.tar.bz2: pingvo-2.3/indexer.cc <==
#include <stdio.h>
#include <stdlib.h>

#ifdef __linux__
#include <unistd.h>
#include <sys/mman.h>
#else
#undef UNICODE
#include <windows.h>
#endif

#include <fcntl.h>
#include <string.h>
#include <q3cstring.h>
#include <qdom.h>
#include <qfile.h>
#include "indexer.h"

#include <algorithm>
using namespace std;

#ifndef __linux__
#include <stdarg.h>
int asprintf( char **s, char *fmt, ... ) {
	int rc, len = 1024;
	char *buf;
	va_list vl;
	
	va_start( vl, fmt );

	buf = (char*) malloc( len );
	for (;;) {
		rc = _vsnprintf( buf, len-1, fmt, vl );
		if ( rc >= 0 ) break;
		len += 1024;
		buf = (char*) realloc( buf, len );
	}
	*s = buf;
	return rc;
}
#define strncasecmp _strnicmp
#endif

char *IndexBuilder::words_start = 0;

IndexBuilder::IndexBuilder(): acount(0) {
  rootNode.word_number = -1;
  lookupTable = &rootNode.sub;
}

void IndexBuilder::build( const char *xmlname ) {
  char *base, *idxname, *wordname, *treename;
  char *idxname2, *wordname2;
  size_t len;
  FILE *f;
  
  base = strdup( xmlname );
  len = strlen(base) - 4 ;
  base[len] = 0;
  asprintf( &idxname, "%s.idx", base );
  asprintf( &wordname, "%s.words", base );
  asprintf( &treename, "%s.tree", base );

  puts( "indexing source xml" );
  build_index( xmlname, idxname );
  printf( "%d articles indexed\n", acount );
  puts( "building wordlist" );
  build_wordlist( xmlname, idxname, wordname );

  puts( "sorting wordlist" );
  asprintf( &idxname2, "%s.idx2", base );
  asprintf( &wordname2, "%s.words2", base );
  sort_words( idxname, idxname2, wordname );
  unlink( idxname );
  rename( idxname2, idxname );

  puts( "writing suffix tree" );
  f = fopen( treename, "wb" );
  if ( !f ) return;
//  ftruncate( fileno(f), 0 );
  write_tree( f, &rootNode );
  fclose( f );
}

void IndexBuilder::addToLookupTable( QString text, int word_num ) {
  int i;
  unsigned j;
  size_t sz;
  QChar c;
  vector <lookupEntry*> *dir;
  lookupEntry *ent=0;

  dir = lookupTable;
  for ( i = 0; i < text.length(); i++ ) {
        c = text.at(i);
        sz = dir->size();
        for ( j = 0; j < sz; j++ )
                if ( (*dir)[j]->c == c ) {
                        ent = (*dir)[j];
                        dir = &ent->sub;
                        break;
                }
        if ( j == sz ) break;
  }

  if ( i == text.length() ) {
        if ( ent ) {
		ent->word_number = word_num;
		ent->word = text;
	}
        return;
  }

  for ( ; i < text.length(); i++ ) {
        ent = new lookupEntry;
        ent->c = text.at(i);
		ent->word_number = -1;
        ent->word = QString::null;
        dir->push_back( ent );
        dir = &ent->sub;
  }

  ent->word_number = word_num;
  ent->word = text;
}

int IndexBuilder::build_index( const char *xmlname, const char *idxname ) {
  int state, offs, loffs, asize;
  unsigned int bidx;
  char buf[80];
  int c;
  FILE *f, *f2;

  f = fopen( xmlname, "rb" );
  if ( !f ) {
  	perror( "build_index fopen(xmlname) failed" );
	return -1;
  }

  f2 = fopen( idxname, "wb" );
  if ( !f2 ) {
  	perror( "build_index fopen(idxname) failed" );
	return -1;
  }

  state = 0;
  acount = 0;
  offs = 0;
  while ( !feof(f) ) {
  	c = fgetc( f );
	if ( c == EOF  ) break;
	switch ( state ) {
	case 0:
		if ( c == '<' ) {
			loffs = offs;
			state = 1;
			bidx = 0;
		}
		break;
	case 1:	
		if ( c != '>' ) {
			if ( bidx < sizeof buf ) buf[bidx++] = c;
			break;
		}

		if ( !strncasecmp( buf, "entry", 5 ) ) state = 2;
		else state = 0;

		break;
	case 2:
		if ( c == '<' ) {
			state = 3;
			bidx = 0;
		}
		break;
	case 3:
		if ( c != '>' ) {
			if ( bidx < sizeof buf ) buf[bidx++] = c;
			break;
		}
		
		if ( strncasecmp( buf, "/entry", 6 ) ) {
			state = 2;
			break;
		}

		state = 0;
		asize = offs - loffs + 1;
		fwrite( &loffs, sizeof offs, 1, f2 );
		fwrite( &asize, sizeof asize, 1, f2 );
		acount++;
		break;
	}
	offs++;
  }
  fclose( f2 );
  fclose( f );

  return acount;
}

void IndexBuilder::build_wordlist( const char *xmlfile,
	const char *idxfile, const char *wordsfile ) {
  QDomDocument doc("entry");
  FILE *idx, *xml, *f2;
  size_t rc;
  int idxrec[2], word_num, idxpos;
  char *buf = 0;
  char bufsize = 0;
  word_desc wdesc;
  QString it_name;

  xml = fopen( xmlfile, "rb" );
  if ( !xml ) return;
  idx = fopen( idxfile, "rb" );
  if ( !idx ) {
  	fclose( xml );
  	return;
  }
  f2 = fopen( wordsfile, "wb" );
  if ( !f2 ) {
  	fclose( idx );
	fclose( xml );
	return;
  }

  fseek( f2, acount * 4, SEEK_SET );
  
  word_num = 0;
  words.clear();

  idxpos = 0;
  for (;;) {
  	char *word;

  	rc = fread( idxrec, 1, sizeof idxrec, idx );
	if ( rc != sizeof idxrec ) break;
	fseek( xml, idxrec[0], SEEK_SET );
	if ( idxrec[1] > bufsize ) {
		buf = (char*) realloc( buf, idxrec[1] );
		bufsize = idxrec[1];
	}
	fread( buf, 1, idxrec[1], xml );
	doc.setContent( Q3CString(buf) );
	it_name = doc.documentElement().attribute("word").simplifyWhiteSpace();
	word = strdup( it_name.utf8().data() );
	wdesc.dict_pos = word_num;
	wdesc.offs = ftell( f2 );
	words.push_back( wdesc );

	fseek( f2, word_num * 4, SEEK_SET );
	fwrite( &wdesc.offs, sizeof wdesc.offs, 1, f2 );
	fseek( f2, wdesc.offs, SEEK_SET );

	fwrite( word, 1, strlen(word)+1, f2 );
	free( word );

	word_num++;
  }

  if ( buf ) free( buf );
  fclose( f2 );
  fclose( idx );
  fclose( xml );
}

bool IndexBuilder::less_worddesc( const word_desc &w1, const word_desc &w2 ) {
  QString s1, s2;
  
  s1 = QString::fromUtf8( words_start + w1.offs );
  s2 = QString::fromUtf8( words_start + w2.offs );

  return s1 < s2;
}

#ifdef __linux__
void IndexBuilder::map_words( char *fname, char* &words_start, int& lenwords ) {
  int fdwords = open( fname, O_RDWR );
  lenwords = lseek( fdwords, 0, SEEK_END );
  words_start = (char*) mmap( 0, lenwords, PROT_READ|PROT_WRITE,
  	MAP_SHARED, fdwords, 0 );
  ::close( fdwords );
}

void IndexBuilder::unmap_words( char *words_start, int lenwords ) {
  munmap( words_start, lenwords );
}

void IndexBuilder::process_idx( char *idxfile, char *idxfile2 ) {
  int fdidx, lenidx, *idx_start;
  unsigned int i;

  fdidx = open( idxfile, O_RDONLY );
  lenidx = lseek( fdidx, 0, SEEK_END );
  idx_start = (int*) mmap( 0, lenidx, PROT_READ, MAP_SHARED, fdidx, 0 );
  close( fdidx );
  fdidx = open( idxfile2, O_RDWR|O_CREAT|O_TRUNC, 0644 );
  for ( i = 0; i < words.size(); i++ )
	write( fdidx, idx_start+(words[i].dict_pos << 1), sizeof(int) * 2 );
  munmap( idx_start, lenidx );
  close( fdidx );
}

#else

void IndexBuilder::map_words( char *file, char* &words_start, int& lenwords ) {
  HANDLE hFile, hMap;
  hFile = CreateFile( file, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
  if ( hFile == INVALID_HANDLE_VALUE ) {
	  puts( "aaa" );
  }
  lenwords = GetFileSize( hFile, NULL );
  hMap = CreateFileMapping( hFile, 0, PAGE_READWRITE, 0, 0, NULL );
  words_start = (char*) MapViewOfFile( hMap, FILE_MAP_READ|FILE_MAP_WRITE, 0, 0, 0 ); 
  CloseHandle( hFile );
  CloseHandle( hMap );
}

void IndexBuilder::unmap_words( char *words_start, int /* lenwords */ ) {
  UnmapViewOfFile( words_start );
}

void IndexBuilder::process_idx( char *idxfile, char *idxfile2 ) {
  HANDLE hFile, hMap;
  int *idx_start;
  unsigned int i;
  DWORD written;

  hFile = CreateFile( idxfile, GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
  hMap = CreateFileMapping( hFile, 0, PAGE_READONLY, 0, 0, NULL );
  idx_start = (int*) MapViewOfFile( hMap, FILE_MAP_READ, 0, 0, 0 );
  CloseHandle( hFile );
  CloseHandle( hMap );

  hFile = CreateFile( idxfile2, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
		NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
  for ( i = 0; i < words.size(); i++ )
	WriteFile( hFile, idx_start+(words[i].dict_pos << 1), sizeof(int) * 2, &written, 0 );
  CloseHandle( hFile );

  UnmapViewOfFile( idx_start );
}
#endif

void IndexBuilder::sort_words( char* idxfile, char* idxfile2,
	char* wordsfile ) {
  unsigned int i;
  int lenwords;
  QString word;

  map_words( wordsfile, words_start, lenwords );
  sort( words.begin(), words.end(), less_worddesc );
  process_idx( idxfile, idxfile2 );

  for ( i = 0; i < words.size(); i++ )
  	((int*)words_start)[i] = words[i].offs;

  puts( "building suffix tree" );
  for ( unsigned int word_num = 0; word_num < words.size(); word_num++ ) {
	word = QString::fromUtf8( words_start + words[word_num].offs );
	addToLookupTable( word, word_num );
  }

  unmap_words( words_start, lenwords );
}

void IndexBuilder::write_tree( FILE *f, lookupEntry *ent ) {
  unsigned int i;
  size_t subpos;
  int pos;
  lookupEntryExt ext;

  ext.c = ent->c.unicode();
  ext.word_number = ent->word_number;
  ext.dchilds_count = ent->sub.size();
  ext.childs_count = ent->childs_count;
  fwrite( &ext, sizeof ext, 1, f );
  pos = ftell( f );

  subpos = pos + ent->sub.size() * 4;
  fseek( f, subpos, SEEK_SET );
  for ( i = 0; i < ent->sub.size(); i++ ) {
	write_tree( f, ent->sub[i] );
	fseek( f, pos + i * 4, SEEK_SET );
	fwrite( &subpos, 4, 1, f );
	fseek( f, 0, SEEK_END );
	subpos = ftell( f );
  }
}

IndexBuilderDsl::IndexBuilderDsl() {
}

void IndexBuilderDsl::build( const char *datname0 ) {
  size_t namelen = strlen( datname0 );
  if ( namelen < 4 || strcmp( datname0 + namelen - 4, ".dat" ) ) return;

  char *base, *idxname2, *treename;
	  
  base = strdup( datname0 );
  base[namelen-4] = 0;

  dsldat = strdup( datname0 );
  asprintf( &dslidx, "%s.ddx", base );
  asprintf( &idxname, "%s.idx", base );
  asprintf( &idxname2, "%s.idx2", base );
  asprintf( &wordsname, "%s.words", base );
  asprintf( &treename, "%s.tree", base );

  acount = build_index();
  build_wordlist();

  sort_words( idxname, idxname2, wordsname );
  unlink( idxname );
  rename( idxname2, idxname );

  FILE *f = fopen( treename, "wb" );
  if ( !f ) return;
  write_tree( f, &rootNode );
  fclose( f );
}

int IndexBuilderDsl::build_index() {
  unsigned short reclen, asize16;
  unsigned char wordlen;
  unsigned int offs, loffs, asize;
  int acount, rc;
  FILE *f, *f1, *f2;

  f = fopen( dslidx, "rb" );
  if ( !f ) {
  	perror( "build_index fopen(dslidx) failed" );
	return -1;
  }

  f1 = fopen( dsldat, "rb" );
  if ( !f1 ) {
	perror( "build_index fopen(dsldat) failed" );
	return -1;
  }

  f2 = fopen( idxname, "wb+" );
  if ( !f2 ) {
  	perror( "build_index fopen(idxname) failed" );
	return -1;
  }

  offs = 0;
  acount = 0;
  while ( !feof(f) ) {
	rc = fread( &reclen, sizeof reclen, 1, f );
	if ( rc != 1 ) break;

	fread( &wordlen, sizeof wordlen, 1, f );
	wordlen <<= 1;
	offs += sizeof(reclen) + sizeof(wordlen) + wordlen;
	fseek( f, offs, SEEK_SET );
	fread( &loffs, sizeof loffs, 1, f );

	fseek( f1, loffs, SEEK_SET );
	fread( &asize16, sizeof asize16, 1, f1 );
	asize = asize16 * 2;

	loffs += sizeof(asize16);
	fwrite( &loffs, sizeof loffs, 1, f2 );
	fwrite( &asize, sizeof asize, 1, f2 );

	offs += sizeof(loffs) + 10;
	fseek( f, offs, SEEK_SET );
	acount++;
  }

  fclose( f );
  fclose( f1 );
  fclose( f2 );

  return acount;
}

void IndexBuilderDsl::build_wordlist() {
  FILE *f, *f1;
  int i, wlen2;
  unsigned int offs, woffs;
  unsigned short reclen;
  unsigned char wordlen;
  unsigned short buf[256];
  QString word;
  Q3CString wordUtf8;
  word_desc wdesc;

  f = fopen( dslidx, "rb" );
  f1 = fopen( wordsname, "wb" );

  words.clear();

  offs = 0;
  woffs = acount * sizeof woffs;
  for ( i = 0; i < acount; i++ ) {
	wdesc.dict_pos = i;
	wdesc.offs = woffs;
	words.push_back( wdesc );

	fseek( f1, i * sizeof woffs, SEEK_SET );
	fwrite( &woffs, sizeof woffs, 1, f1 );

	fseek( f1, woffs, SEEK_SET );

	fread( &reclen, sizeof reclen, 1, f );
	fread( &wordlen, sizeof wordlen, 1, f );
	wlen2 = wordlen << 1;
	fread( buf, wlen2, 1, f );
	buf[wordlen] = 0;
	word = QString::fromUcs2( buf );
	wordUtf8 = word.utf8();
	fwrite( wordUtf8.data(), wordUtf8.length()+1, 1, f1 );
	woffs += wordUtf8.length() + 1;

	offs += sizeof(reclen) + sizeof(wordlen) + wlen2 + 4 + 10;
	fseek( f, offs, SEEK_SET );
  }

  fclose( f );
  fclose( f1 );
}
